# GtkSpellCheckerDialog for pyenchant
#
# Copyright (C) 2004-2005, Fredrik Corneliusson
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# In addition, as a special exception, you are
# given permission to link the code of this program with
# non-LGPL Spelling Provider libraries (eg: a MSFT Office
# spell checker backend) and distribute linked combinations including
# the two.  You must obey the GNU Lesser General Public License in all
# respects for all of the code used other than said providers.  If you modify
# this file, you may extend this exception to your version of the
# file, but you are not obligated to do so.  If you do not wish to
# do so, delete this exception statement from your version.
#

import gtk


#   columns
COLUMN_SUGGESTION = 0


def create_list_view(
    col_label,
):
    # create list widget
    list_ = gtk.ListStore(str)
    list_view = gtk.TreeView(model=list_)

    list_view.set_rules_hint(True)
    list_view.get_selection().set_mode(gtk.SELECTION_SINGLE)
    # Add Colums
    renderer = gtk.CellRendererText()
    renderer.set_data("column", COLUMN_SUGGESTION)
    column = gtk.TreeViewColumn(col_label, renderer, text=COLUMN_SUGGESTION)
    list_view.append_column(column)
    return list_view


class GtkSpellCheckerDialog(gtk.Window):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_title("Spell check")
        self.set_default_size(350, 200)

        self._checker = None
        self._numContext = 40

        self.errors = None

        # create accel group
        accel_group = gtk.AccelGroup()
        self.add_accel_group(accel_group)

        # list of widgets to disable if there's no spell error left
        self._conditional_widgets = []
        conditional = self._conditional_widgets.append

        # layout
        mainbox = gtk.VBox(spacing=5)
        hbox = gtk.HBox(spacing=5)
        self.add(mainbox)
        mainbox.pack_start(hbox, padding=5)

        box1 = gtk.VBox(spacing=5)
        hbox.pack_start(box1, padding=5)
        conditional(box1)

        # unrecognized word
        text_view_lable = gtk.Label("Unrecognized word")
        text_view_lable.set_justify(gtk.JUSTIFY_LEFT)
        box1.pack_start(text_view_lable, False, False)

        text_view = gtk.TextView()
        text_view.set_wrap_mode(gtk.WRAP_WORD)
        text_view.set_editable(False)
        text_view.set_cursor_visible(False)
        self.error_text = text_view.get_buffer()
        text_buffer = text_view.get_buffer()
        text_buffer.create_tag("fg_black", foreground="black")
        text_buffer.create_tag("fg_red", foreground="red")

        box1.pack_start(text_view)

        # Change to
        change_to_box = gtk.HBox()
        box1.pack_start(change_to_box, False, False)

        change_to_label = gtk.Label("Change to:")
        self.replace_text = gtk.Entry()
        text_view_lable.set_justify(gtk.JUSTIFY_LEFT)
        change_to_box.pack_start(change_to_label, False, False)
        change_to_box.pack_start(self.replace_text)

        # scrolled window
        sw = gtk.ScrolledWindow()
        sw.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        box1.pack_start(sw)

        self.suggestion_list_view = create_list_view("Suggestions")
        self.suggestion_list_view.connect("button_press_event", self._onButtonPress)
        self.suggestion_list_view.connect("cursor-changed", self._onSuggestionChanged)
        sw.add(self.suggestion_list_view)

        # ---Buttons---#000000#FFFFFF----------------------------------------------------
        button_box = gtk.VButtonBox()
        hbox.pack_start(button_box, False, False)

        # Ignore
        button = gtk.Button("Ignore")
        button.connect("clicked", self._onIgnore)
        button.add_accelerator(
            "activate", accel_group, gtk.keysyms.Return, 0, gtk.ACCEL_VISIBLE
        )
        button_box.pack_start(button)
        conditional(button)

        # Ignore all
        button = gtk.Button("Ignore All")
        button.connect("clicked", self._onIgnoreAll)
        button_box.pack_start(button)
        conditional(button)

        # Replace
        button = gtk.Button("Replace")
        button.connect("clicked", self._onReplace)
        button_box.pack_start(button)
        conditional(button)

        # Replace all
        button = gtk.Button("Replace All")
        button.connect("clicked", self._onReplaceAll)
        button_box.pack_start(button)
        conditional(button)

        # Recheck button
        button = gtk.Button("_Add")
        button.connect("clicked", self._onAdd)

        button_box.pack_start(button)
        conditional(button)

        # Close button
        button = gtk.Button(stock=gtk.STOCK_CLOSE)
        button.connect("clicked", self._onClose)
        button.add_accelerator(
            "activate", accel_group, gtk.keysyms.Escape, 0, gtk.ACCEL_VISIBLE
        )
        button_box.pack_end(button)

        # dictionary label
        self._dict_lable = gtk.Label("")
        mainbox.pack_start(self._dict_lable, False, False, padding=5)

        mainbox.show_all()

    def _onIgnore(self, w, *args):
        print(["ignore"])
        self._advance()

    def _onIgnoreAll(self, w, *args):
        print(["ignore all"])
        self._checker.ignore_always()
        self._advance()

    def _onReplace(self, *args):
        print(["Replace"])
        repl = self._getRepl()
        self._checker.replace(repl)
        self._advance()

    def _onReplaceAll(self, *args):
        print(["Replace all"])
        repl = self._getRepl()
        self._checker.replace_always(repl)
        self._advance()

    def _onAdd(self, *args):
        """Callback for the "add" button."""
        self._checker.add()
        self._advance()

    def _onClose(self, w, *args):
        self.emit("delete_event", gtk.gdk.Event(gtk.gdk.BUTTON_PRESS))
        return True

    def _onButtonPress(self, widget, event):
        if event.type == gtk.gdk._2BUTTON_PRESS:
            print(["Double click!"])
            self._onReplace()

    def _onSuggestionChanged(self, widget, *args):
        selection = self.suggestion_list_view.get_selection()
        model, iter = selection.get_selected()
        if iter:
            suggestion = model.get_value(iter, COLUMN_SUGGESTION)
            self.replace_text.set_text(suggestion)

    def _getRepl(self):
        """Get the chosen replacement string."""
        repl = self.replace_text.get_text()
        repl = self._checker.coerce_string(repl)
        return repl

    def _fillSuggestionList(self, suggestions):
        model = self.suggestion_list_view.get_model()
        model.clear()
        for suggestion in suggestions:
            value = "%s" % (suggestion,)
            model.append(
                [
                    value,
                ]
            )

    def setSpellChecker(self, checker):
        assert checker, "checker can't be None"
        self._checker = checker
        self._dict_lable.set_text("Dictionary:%s" % (checker.dict.tag,))

    def getSpellChecker(self, checker):
        return self._checker

    def updateUI(self):
        self._advance()

    def _disableButtons(self):
        for w in self._conditional_widgets:
            w.set_sensitive(False)

    def _enableButtons(self):
        for w in self._conditional_widgets:
            w.set_sensitive(True)

    def _advance(self):
        """Advance to the next error.
        This method advances the SpellChecker to the next error, if
        any.  It then displays the error and some surrounding context,
        and well as listing the suggested replacements.
        """
        # Disable interaction if no checker
        if self._checker is None:
            self._disableButtons()
            self.emit("check-done")
            return

        # Advance to next error, disable if not available
        try:
            self._checker.next()
        except StopIteration:
            self._disableButtons()
            self.error_text.set_text("")
            self._fillSuggestionList([])
            self.replace_text.set_text("")
            return
        self._enableButtons()

        # Display error context with erroneous word in red
        self.error_text.set_text("")
        iter = self.error_text.get_iter_at_offset(0)
        append = self.error_text.insert_with_tags_by_name

        lContext = self._checker.leading_context(self._numContext)
        tContext = self._checker.trailing_context(self._numContext)
        append(iter, lContext, "fg_black")
        append(iter, self._checker.word, "fg_red")
        append(iter, tContext, "fg_black")

        # Display suggestions in the replacements list
        suggs = self._checker.suggest()
        self._fillSuggestionList(suggs)
        if suggs:
            self.replace_text.set_text(suggs[0])
        else:
            self.replace_text.set_text("")


def _test():
    from enchant.checker import SpellChecker

    text = "This is sme text with a fw speling errors in it. Here are a fw more to tst it ut."
    print(["BEFORE:", text])
    chk_dlg = GtkSpellCheckerDialog()
    chk_dlg.show()
    chk_dlg.connect("delete_event", gtk.main_quit)

    chkr = SpellChecker("en_US", text)

    chk_dlg.setSpellChecker(chkr)
    chk_dlg.updateUI()
    gtk.main()


if __name__ == "__main__":
    _test()



# pyenchant
#
# Copyright (C) 2004-2008 Ryan Kelly
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# In addition, as a special exception, you are
# given permission to link the code of this program with
# non-LGPL Spelling Provider libraries (eg: a MSFT Office
# spell checker backend) and distribute linked combinations including
# the two.  You must obey the GNU Lesser General Public License in all
# respects for all of the code used other than said providers.  If you modify
# this file, you may extend this exception to your version of the
# file, but you are not obligated to do so.  If you do not wish to
# do so, delete this exception statement from your version.
#
"""

enchant.utils:    Misc utilities for the enchant package
========================================================

This module provides miscellaneous utilities for use with the
enchant spellchecking package.  Currently available functionality
includes:

    * functions for dealing with locale/language settings
    * ability to list supporting data files (win32 only)
    * functions for bundling supporting data files from a build

"""

from enchant.errors import Error
import locale


def levenshtein(s1, s2):
    """Calculate the Levenshtein distance between two strings.

    This is straight from Wikipedia.
    """
    if len(s1) < len(s2):
        return levenshtein(s2, s1)
    if not s1:
        return len(s2)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def trim_suggestions(word, suggs, maxlen, calcdist=None):
    """Trim a list of suggestions to a maximum length.

    If the list of suggested words is too long, you can use this function
    to trim it down to a maximum length.  It tries to keep the "best"
    suggestions based on similarity to the original word.

    If the optional "calcdist" argument is provided, it must be a callable
    taking two words and returning the distance between them.  It will be
    used to determine which words to retain in the list.  The default is
    a simple Levenshtein distance.
    """
    if calcdist is None:
        calcdist = levenshtein
    decorated = [(calcdist(word, s), s) for s in suggs]
    decorated.sort()
    return [s for (l, s) in decorated[:maxlen]]


def get_default_language(default=None):
    """Determine the user's default language, if possible.

    This function uses the 'locale' module to try to determine
    the user's preferred language.  The return value is as
    follows:

        * if a locale is available for the LC_MESSAGES category,
          that language is used
        * if a default locale is available, that language is used
        * if the keyword argument <default> is given, it is used
        * if nothing else works, None is returned

    Note that determining the user's language is in general only
    possible if they have set the necessary environment variables
    on their system.
    """
    try:
        tag = locale.getlocale()[0]
        if tag is None:
            tag = locale.getdefaultlocale()[0]
            if tag is None:
                raise Error("No default language available")
        return tag
    except Exception:
        pass
    return default


get_default_language._DOC_ERRORS = ["LC"]



# pyenchant
#
# Copyright (C) 2004-2008, Ryan Kelly
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# In addition, as a special exception, you are
# given permission to link the code of this program with
# non-LGPL Spelling Provider libraries (eg: a MSFT Office
# spell checker backend) and distribute linked combinations including
# the two.  You must obey the GNU Lesser General Public License in all
# respects for all of the code used other than said providers.  If you modify
# this file, you may extend this exception to your version of the
# file, but you are not obligated to do so.  If you do not wish to
# do so, delete this exception statement from your version.
#
"""

enchant.checker:  High-level spellchecking functionality
========================================================

This package is designed to host higher-level spellchecking functionality
than is available in the base enchant package.  It should make writing
applications that follow common usage idioms significantly easier.

The most useful class is SpellChecker, which implements a spellchecking
loop over a block of text.  It is capable of modifying the text in-place
if given an array of characters to work with.

This package also contains several interfaces to the SpellChecker class,
such as a wxPython GUI dialog and a command-line interface.

"""

import array
import warnings

import enchant
from enchant.errors import (
    DefaultLanguageNotFoundError,
    DictNotFoundError,
    TokenizerNotFoundError,
)
from enchant.tokenize import get_tokenizer
from enchant.utils import get_default_language


class SpellChecker:
    """Class implementing stateful spellchecking behaviour.

    This class is designed to implement a spell-checking loop over
    a block of text, correcting/ignoring/replacing words as required.
    This loop is implemented using an iterator paradigm so it can be
    embedded inside other loops of control.

    The SpellChecker object is stateful, and the appropriate methods
    must be called to alter its state and affect the progress of
    the spell checking session.  At any point during the checking
    session, the attribute 'word' will hold the current erroneously
    spelled word under consideration.  The action to take on this word
    is determined by calling methods such as 'replace', 'replace_always'
    and 'ignore_always'.  Once this is done, calling 'next' advances
    to the next misspelled word.

    As a quick (and rather silly) example, the following code replaces
    each misspelled word with the string "SPAM":

        >>> text = "This is sme text with a fw speling errors in it."
        >>> chkr = SpellChecker("en_US",text)
        >>> for err in chkr:
        ...   err.replace("SPAM")
        ...
        >>> chkr.get_text()
        'This is SPAM text with a SPAM SPAM errors in it.'
        >>>

    Internally, the SpellChecker always works with arrays of (possibly
    unicode) character elements.  This allows the in-place modification
    of the string as it is checked, and is the closest thing Python has
    to a mutable string.  The text can be set as any of a normal string,
    unicode string, character array or unicode character array. The
    'get_text' method will return the modified array object if an
    array is used, or a new string object if a string it used.

    Words input to the SpellChecker may be either plain strings or
    unicode objects.  They will be converted to the same type as the
    text being checked, using python's default encoding/decoding
    settings.

    If using an array of characters with this object and the
    array is modified outside of the spellchecking loop, use the
    'set_offset' method to reposition the internal loop pointer
    to make sure it doesn't skip any words.

    """

    _DOC_ERRORS = ["sme", "fw", "speling", "chkr", "chkr", "chkr"]

    def __init__(
        self, lang=None, text=None, tokenize=None, chunkers=None, filters=None
    ):
        """Constructor for the SpellChecker class.

        SpellChecker objects can be created in two ways, depending on
        the nature of the first argument.  If it is a string, it
        specifies a language tag from which a dictionary is created.
        Otherwise, it must be an enchant Dict object to be used.

        Optional keyword arguments are:

            * text:  to set the text to be checked at creation time
            * tokenize:  a custom tokenization function to use
            * chunkers:  a list of chunkers to apply during tokenization
            * filters:  a list of filters to apply during tokenization

        If <tokenize> is not given and the first argument is a Dict,
        its 'tag' attribute must be a language tag so that a tokenization
        function can be created automatically.  If this attribute is missing
        the user's default language will be used.
        """
        if lang is None:
            lang = get_default_language()
        if isinstance(lang, (str, bytes)):
            try:
                dict = enchant.Dict(lang)
            except DictNotFoundError:
                raise DefaultLanguageNotFoundError(lang) from None
        else:
            dict = lang
            try:
                lang = dict.tag
            except AttributeError:
                lang = get_default_language()
        if lang is None:
            raise DefaultLanguageNotFoundError from None
        self.lang = lang
        self.dict = dict
        if tokenize is None:
            try:
                tokenize = get_tokenizer(lang, chunkers, filters)
            except TokenizerNotFoundError:
                # Fall back to default tokenization if no match for 'lang'
                tokenize = get_tokenizer(None, chunkers, filters)
        self._tokenize = tokenize

        self.word = None
        self.wordpos = None
        self._ignore_words = {}
        self._replace_words = {}
        # Default to the empty string as the text to be checked
        self._text = array.array("u")
        self._use_tostring = False
        self._tokens = iter([])

        if text is not None:
            self.set_text(text)

    def __iter__(self):
        """Each SpellChecker object is its own iterator"""
        return self

    def set_text(self, text):
        """Set the text to be spell-checked.

        This method must be called, or the 'text' argument supplied
        to the constructor, before calling the 'next()' method.
        """
        # Convert to an array object if necessary
        if isinstance(text, (str, bytes)):
            if type(text) is str:
                self._text = array.array("u", text)
            else:
                self._text = array.array("c", text)
            self._use_tostring = True
        else:
            self._text = text
            self._use_tostring = False
        self._tokens = self._tokenize(self._text)

    def get_text(self):
        """Return the spell-checked text."""
        if self._use_tostring:
            return self._array_to_string(self._text)
        return self._text

    def _array_to_string(self, text):
        """Format an internal array as a standard string."""
        if text.typecode == "u":
            return text.tounicode()
        return text.tostring()

    def wants_unicode(self):
        """Check whether the checker wants unicode strings.

        This method will return True if the checker wants unicode strings
        as input, False if it wants normal strings.  It's important to
        provide the correct type of string to the checker.
        """
        if self._text.typecode == "u":
            return True
        return False

    def coerce_string(self, text, enc=None):
        """Coerce string into the required type.

        This method can be used to automatically ensure that strings
        are of the correct type required by this checker - either unicode
        or standard.  If there is a mismatch, conversion is done using
        python's default encoding unless another encoding is specified.
        """
        if self.wants_unicode():
            if not isinstance(text, str):
                if enc is None:
                    return text.decode()
                else:
                    return text.decode(enc)
            return text
        if not isinstance(text, bytes):
            if enc is None:
                return text.encode()
            else:
                return text.encode(enc)
        return text

    def __next__(self):
        return self.next()

    def next(self):
        """Process text up to the next spelling error.

        This method is designed to support the iterator protocol.
        Each time it is called, it will advance the 'word' attribute
        to the next spelling error in the text.  When no more errors
        are found, it will raise StopIteration.

        The method will always return self, so that it can be used
        sensibly in common idioms such as:

            for err in checker:
                err.do_something()

        """
        # Find the next spelling error.
        # The uncaught StopIteration from next(self._tokens)
        # will provide the StopIteration for this method
        while True:
            (word, pos) = next(self._tokens)
            # decode back to a regular string
            word = self._array_to_string(word)
            if self.dict.check(word):
                continue
            if word in self._ignore_words:
                continue
            self.word = word
            self.wordpos = pos
            if word in self._replace_words:
                self.replace(self._replace_words[word])
                continue
            break
        return self

    def replace(self, repl):
        """Replace the current erroneous word with the given string."""
        repl = self.coerce_string(repl)
        a_repl = array.array(self._text.typecode, repl)
        if repl:
            self.dict.store_replacement(self.word, repl)
        self._text[self.wordpos : self.wordpos + len(self.word)] = a_repl
        incr = len(repl) - len(self.word)
        self._tokens.set_offset(self._tokens.offset + incr, replaced=True)

    def replace_always(self, word, repl=None):
        """Always replace given word with given replacement.

        If a single argument is given, this is used to replace the
        current erroneous word.  If two arguments are given, that
        combination is added to the list for future use.
        """
        if repl is None:
            repl = word
            word = self.word
        repl = self.coerce_string(repl)
        word = self.coerce_string(word)
        self._replace_words[word] = repl
        if self.word == word:
            self.replace(repl)

    def ignore_always(self, word=None):
        """Add given word to list of words to ignore.

        If no word is given, the current erroneous word is added.
        """
        if word is None:
            word = self.word
        word = self.coerce_string(word)
        if word not in self._ignore_words:
            self._ignore_words[word] = True

    def add_to_personal(self, word=None):
        """Add given word to the personal word list.

        If no word is given, the current erroneous word is added.
        """
        warnings.warn(
            "SpellChecker.add_to_personal is deprecated, "
            "please use SpellChecker.add",
            category=DeprecationWarning,
            stacklevel=2,
        )
        self.add(word)

    def add(self, word=None):
        """Add given word to the personal word list.

        If no word is given, the current erroneous word is added.
        """
        if word is None:
            word = self.word
        self.dict.add(word)

    def suggest(self, word=None):
        """Return suggested spellings for the given word.

        If no word is given, the current erroneous word is used.
        """
        if word is None:
            word = self.word
        suggs = self.dict.suggest(word)
        return suggs

    def check(self, word):
        """Check correctness of the given word."""
        return self.dict.check(word)

    def set_offset(self, off, whence=0):
        """Set the offset of the tokenization routine.

        For more details on the purpose of the tokenization offset,
        see the documentation of the 'enchant.tokenize' module.
        The optional argument whence indicates the method by
        which to change the offset:

            * 0 (the default) treats <off> as an increment
            * 1 treats <off> as a distance from the start
            * 2 treats <off> as a distance from the end
        """
        if whence == 0:
            self._tokens.set_offset(self._tokens.offset + off)
        elif whence == 1:
            assert off > 0
            self._tokens.set_offset(off)
        elif whence == 2:
            assert off > 0
            self._tokens.set_offset(len(self._text) - 1 - off)
        else:
            raise ValueError("Invalid value for whence: %s" % (whence,))

    def leading_context(self, chars):
        """Get <chars> characters of leading context.

        This method returns up to <chars> characters of leading
        context - the text that occurs in the string immediately
        before the current erroneous word.
        """
        start = max(self.wordpos - chars, 0)
        context = self._text[start : self.wordpos]
        return self._array_to_string(context)

    def trailing_context(self, chars):
        """Get <chars> characters of trailing context.

        This method returns up to <chars> characters of trailing
        context - the text that occurs in the string immediately
        after the current erroneous word.
        """
        start = self.wordpos + len(self.word)
        end = min(start + chars, len(self._text))
        context = self._text[start:end]
        return self._array_to_string(context)



# pyenchant
#
# Copyright (C) 2004-2008, Ryan Kelly
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# In addition, as a special exception, you are
# given permission to link the code of this program with
# non-LGPL Spelling Provider libraries (eg: a MSFT Office
# spell checker backend) and distribute linked combinations including
# the two.  You must obey the GNU Lesser General Public License in all
# respects for all of the code used other than said providers.  If you modify
# this file, you may extend this exception to your version of the
# file, but you are not obligated to do so.  If you do not wish to
# do so, delete this exception statement from your version.
#
"""

    enchant.checker.CmdLineChecker:  Command-Line spell checker

    This module provides the class CmdLineChecker, which interactively
    spellchecks a piece of text by interacting with the user on the
    command line.  It can also be run as a script to spellcheck a file.

"""

import sys

from enchant.checker import SpellChecker

# Helpers

colors = {
    "normal": "\x1b[0m",
    "black": "\x1b[30m",
    "red": "\x1b[31m",
    "green": "\x1b[32m",
    "yellow": "\x1b[33m",
    "blue": "\x1b[34m",
    "purple": "\x1b[35m",
    "cyan": "\x1b[36m",
    "grey": "\x1b[90m",
    "gray": "\x1b[90m",
    "bold": "\x1b[1m",
}


def color(string, color="normal", prefix=""):
    """
    Change text color for the Linux terminal.

    Args:
        string (str): String to colorify
        color (str): Color to colorify the string in the following list:
            black, red, green, yellow, blue, purple, cyan, gr[ae]y
        prefix (str): Prefix to add to string (ex: Beginning of line graphics)
    """
    if sys.stdout.isatty():
        return colors[color] + prefix + string + colors["normal"]
    else:
        return prefix + string


def success(string):
    return "[" + color("+", color="green") + "] " + string


def error(string):
    return "[" + color("!", color="red") + "] " + string


def warning(string):
    return "[" + color("*", color="yellow") + "] " + string


def info(string):
    return "[" + color(".", color="blue") + "] " + string


class CmdLineChecker:
    """A simple command-line spell checker.

    This class implements a simple command-line spell checker.  It must
    be given a SpellChecker instance to operate on, and interacts with
    the user by printing instructions on stdout and reading commands from
    stdin.
    """

    _DOC_ERRORS = ["stdout", "stdin"]

    def __init__(self):
        self._stop = False
        self._checker = None

    def set_checker(self, chkr):
        self._checker = chkr

    def get_checker(self, chkr):
        return self._checker

    def run(self):
        """Run the spellchecking loop."""
        self._stop = False
        for err in self._checker:
            self.error = err
            self.print_error()
            self.print_suggestions()
            status = self.read_command()
            while not status and not self._stop:
                status = self.read_command()
            if self._stop:
                break

    def print_error(self):
        """print the spelling error to the console.

        Prints the misspelled word along with 100 characters of
        context on either side.  This number was arbitrarily chosen
        and could be modified to be tunable or changed entirely.
        It seems to be enough context to be helpful though
        """
        error_string = self._build_context(
            self.error.get_text(), self.error.word, self.error.wordpos
        )
        print([error("ERROR: %s" % color(self.error.word, color="red"))])
        print([info("")])
        print([info(error_string)])
        print([info("")])

    @staticmethod
    def _build_context(text, error_word, error_start):
        """creates the context line.

        This function will search forward and backward
        from the error word to find the nearest newlines.
        it will return this line with the error word
        colored red."""
        start_newline = text.rfind("\n", 0, error_start)
        end_newline = text.find("\n", error_start)
        return text[start_newline + 1 : end_newline].replace(
            error_word, color(error_word, color="red")
        )

    def print_suggestions(self):
        """Prints out the suggestions for a given error.

        This function will add vertical pipes to separate choices
        as well as the index of the replacement as expected by the replace function.
        I don't believe zero indexing is a problem as long as the user can see the numbers :)
        """
        result = ""
        suggestions = self.error.suggest()
        for index, sugg in enumerate(suggestions):
            if index == 0:
                result = (
                    result
                    + color(str(index), color="yellow")
                    + ": "
                    + color(sugg, color="bold")
                )
            else:
                result = (
                    result
                    + " | "
                    + color(str(index), color="yellow")
                    + ": "
                    + color(sugg, color="bold")
                )
        print([info("HOW ABOUT:"), result])

    def print_help(self):
        print(
            [
                info(
                    color("0", color="yellow")
                    + ".."
                    + color("N", color="yellow")
                    + ":\t"
                    + color("replace", color="bold")
                    + " with the numbered suggestion"
                )
            ]
        )
        print(
            [
                info(
                    color("R", color="cyan")
                    + color("0", color="yellow")
                    + ".."
                    + color("R", color="cyan")
                    + color("N", color="yellow")
                    + ":\t"
                    + color("always replace", color="bold")
                    + " with the numbered suggestion"
                )
            ]
        )
        print(
            [
                info(
                    color("i", color="cyan")
                    + ":\t\t"
                    + color("ignore", color="bold")
                    + " this word"
                )
            ]
        )
        print(
            [
                info(
                    color("I", color="cyan")
                    + ":\t\t"
                    + color("always ignore", color="bold")
                    + " this word"
                )
            ]
        )
        print(
            [
                info(
                    color("a", color="cyan")
                    + ":\t\t"
                    + color("add", color="bold")
                    + " word to personal dictionary"
                )
            ]
        )
        print(
            [
                info(
                    color("e", color="cyan")
                    + ":\t\t"
                    + color("edit", color="bold")
                    + " the word"
                )
            ]
        )
        print(
            [
                info(
                    color("q", color="cyan")
                    + ":\t\t"
                    + color("quit", color="bold")
                    + " checking"
                )
            ]
        )
        print(
            [
                info(
                    color("h", color="cyan")
                    + ":\t\tprint this "
                    + color("help", color="bold")
                    + " message"
                )
            ]
        )
        print([info("----------------------------------------------------")])
        self.print_suggestions()

    def read_command(self):
        cmd = input(">> ")
        cmd = cmd.strip()

        if cmd.isdigit():
            repl = int(cmd)
            suggs = self.error.suggest()
            if repl >= len(suggs):
                print([warning("No suggestion number"), repl])
                return False
            print(
                [
                    success(
                        "Replacing '%s' with '%s'"
                        % (
                            color(self.error.word, color="red"),
                            color(suggs[repl], color="green"),
                        )
                    )
                ]
            )
            self.error.replace(suggs[repl])
            return True

        if cmd[0] == "R":
            if not cmd[1:].isdigit():
                print([warning("Badly formatted command (try 'help')")])
                return False
            repl = int(cmd[1:])
            suggs = self.error.suggest()
            if repl >= len(suggs):
                print([warning("No suggestion number"), repl])
                return False
            self.error.replace_always(suggs[repl])
            return True

        if cmd == "i":
            return True

        if cmd == "I":
            self.error.ignore_always()
            return True

        if cmd == "a":
            self.error.add()
            return True

        if cmd == "e":
            repl = input(info("New Word: "))
            self.error.replace(repl.strip())
            return True

        if cmd == "q":
            self._stop = True
            return True

        if "help".startswith(cmd.lower()):
            self.print_help()
            return False

        print([warning("Badly formatted command (try 'help')")])
        return False

    def run_on_file(self, infile, outfile=None, enc=None):
        """Run spellchecking on the named file.
        This method can be used to run the spellchecker over the named file.
        If <outfile> is not given, the corrected contents replace the contents
        of <infile>.  If <outfile> is given, the corrected contents will be
        written to that file.  Use "-" to have the contents written to stdout.
        If <enc> is given, it specifies the encoding used to read the
        file's contents into a unicode string.  The output will be written
        in the same encoding.
        """
        inStr = "".join(open(infile).readlines())
        if enc is not None:
            inStr = inStr.decode(enc)
        self._checker.set_text(inStr)
        begin_msg = "Beginning spell check of %s" % infile
        print([info(begin_msg)])
        print([info("-" * len(begin_msg))])
        self.run()
        print([success("Completed spell check of %s" % infile)])
        outStr = self._checker.get_text()
        if enc is not None:
            outStr = outStr.encode(enc)
        if outfile is None:
            outF = open(infile, "w")
        elif outfile == "-":
            outF = sys.stdout
        else:
            outF = open(outfile, "w")
        outF.write(outStr)
        outF.close()

    run_on_file._DOC_ERRORS = ["outfile", "infile", "outfile", "stdout"]


def _run_as_script():
    """Run the command-line spellchecker as a script.
    This function allows the spellchecker to be invoked from the command-line
    to check spelling in a file.
    """
    # Check necessary command-line options
    from optparse import OptionParser

    op = OptionParser()
    op.add_option(
        "-o", "--output", dest="outfile", metavar="FILE", help="write changes into FILE"
    )
    op.add_option(
        "-l",
        "--lang",
        dest="lang",
        metavar="TAG",
        default="en_US",
        help="use language idenfified by TAG",
    )
    op.add_option(
        "-e",
        "--encoding",
        dest="enc",
        metavar="ENC",
        help="file is unicode with encoding ENC",
    )
    (opts, args) = op.parse_args()
    # Sanity check
    if len(args) < 1:
        raise ValueError("Must name a file to check")
    if len(args) > 1:
        raise ValueError("Can only check a single file")
    # Create and run the checker
    chkr = SpellChecker(opts.lang)
    cmdln = CmdLineChecker()
    cmdln.set_checker(chkr)
    cmdln.run_on_file(args[0], opts.outfile, opts.enc)


if __name__ == "__main__":
    _run_as_script()



# pyenchant
#
# Copyright (C) 2004-2009, Ryan Kelly
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#
# In addition, as a special exception, you are
# given permission to link the code of this program with
# non-LGPL Spelling Provider libraries (eg: a MSFT Office
# spell checker backend) and distribute linked combinations including
# the two.  You must obey the GNU Lesser General Public License in all
# respects for all of the code used other than said providers.  If you modify
# this file, you may extend this exception to your version of the
# file, but you are not obligated to do so.  If you do not wish to
# do so, delete this exception statement from your version.
#
"""

enchant.tokenize:    String tokenization functions for PyEnchant
================================================================

An important task in spellchecking is breaking up large bodies of
text into their constituent words, each of which is then checked
for correctness.  This package provides Python functions to split
strings into words according to the rules of a particular language.

Each tokenization function accepts a string as its only positional
argument, and returns an iterator that yields tuples of the following
form, one for each word found::

    (<word>,<pos>)

The meanings of these fields should be clear: <word> is the word
that was found and <pos> is the position within the text at which
the word began (zero indexed, of course).  The function will work
on any string-like object that supports array-slicing; in particular
character-array objects from the 'array' module may be used.

The iterator also provides the attribute 'offset' which gives the current
position of the tokenizer inside the string being split, and the method
'set_offset' for manually adjusting this position.  This can be used for
example if the string's contents have changed during the tokenization
process.

To obtain an appropriate tokenization function for the language
identified by <tag>, use the function 'get_tokenizer(tag)'::

    tknzr = get_tokenizer("en_US")
    for (word,pos) in tknzr("text to be tokenized goes here")
        do_something(word)

This library is designed to be easily extendible by third-party
authors.  To register a tokenization function for the language
<tag>, implement it as the function 'tokenize' within the
module enchant.tokenize.<tag>.  The 'get_tokenizer' function
will automatically detect it.  Note that the underscore must be
used as the tag component separator in this case, in order to
form a valid python module name. (e.g. "en_US" rather than "en-US")

Currently, a tokenizer has only been implemented for the English
language.  Based on the author's limited experience, this should
be at least partially suitable for other languages.

This module also provides various implementations of "Chunkers" and
"Filters".  These classes are designed to make it easy to work with
text in a variety of common formats, by detecting and excluding parts
of the text that don't need to be checked.

A Chunker is a class designed to break a body of text into large chunks
of checkable content; for example the HTMLChunker class extracts the
text content from all HTML tags but excludes the tags themselves.
A Filter is a class designed to skip individual words during the checking
process; for example the URLFilter class skips over any words that
have the format of a URL.

For example, to spellcheck an HTML document it is necessary to split the
text into chunks based on HTML tags, and to filter out common word forms
such as URLs and WikiWords.  This would look something like the following::

    tknzr = get_tokenizer("en_US",(HTMLChunker,),(URLFilter,WikiWordFilter)))

    text = "<html><body>the url is http://example.com</body></html>"
    for (word,pos) in tknzer(text):
        ...check each word and react accordingly...

"""
_DOC_ERRORS = [
    "pos",
    "pos",
    "tknzr",
    "URLFilter",
    "WikiWordFilter",
    "tkns",
    "tknzr",
    "pos",
    "tkns",
]

import re
import warnings
import array

from enchant.errors import TokenizerNotFoundError

#  For backwards-compatibility.  This will eventually be removed, but how
#  does one mark a module-level constant as deprecated?
Error = TokenizerNotFoundError


class tokenize:  # noqa: N801
    """Base class for all tokenizer objects.

    Each tokenizer must be an iterator and provide the 'offset'
    attribute as described in the documentation for this module.

    While tokenizers are in fact classes, they should be treated
    like functions, and so are named using lower_case rather than
    the CamelCase more traditional of class names.
    """

    _DOC_ERRORS = ["CamelCase"]

    def __init__(self, text):
        self._text = text
        self._offset = 0

    def __next__(self):
        return self.next()

    def next(self):
        raise NotImplementedError()

    def __iter__(self):
        return self

    def set_offset(self, offset, replaced=False):
        self._offset = offset

    def _get_offset(self):
        return self._offset

    def _set_offset(self, offset):
        msg = (
            "changing a tokenizers 'offset' attribute is deprecated;"
            " use the 'set_offset' method"
        )
        warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
        self.set_offset(offset)

    offset = property(_get_offset, _set_offset)


def get_tokenizer(tag=None, chunkers=None, filters=None):
    """Locate an appropriate tokenizer by language tag.

    This requires importing the function 'tokenize' from an appropriate
    module.  Modules tried are named after the language tag, tried in the
    following order:

        * the entire tag (e.g. "en_AU.py")
        * the base country code of the tag (e.g. "en.py")

    If the language tag is None, a default tokenizer (actually the English
    one) is returned.  It's unicode aware and should work OK for most
    latin-derived languages.

    If a suitable function cannot be found, raises TokenizerNotFoundError.

    If given and not None, 'chunkers' and 'filters' must be lists of chunker
    classes and filter classes respectively.  These will be applied to the
    tokenizer during creation.
    """
    if tag is None:
        tag = "en"
    # "filters" used to be the second argument.  Try to catch cases
    # where it is given positionally and issue a DeprecationWarning.
    if chunkers is not None and filters is None:
        chunkers = list(chunkers)
        if chunkers:
            try:
                chunkers_are_filters = issubclass(chunkers[0], Filter)
            except TypeError:
                pass
            else:
                if chunkers_are_filters:
                    msg = (
                        "passing 'filters' as a non-keyword argument "
                        "to get_tokenizer() is deprecated"
                    )
                    warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
                    filters = chunkers
                    chunkers = None
    # Ensure only '_' used as separator
    tag = tag.replace("-", "_")
    # First try the whole tag
    tk_func = _try_tokenizer(tag)
    if tk_func is None:
        # Try just the base
        base = tag.split("_")[0]
        tk_func = _try_tokenizer(base)
        if tk_func is None:
            msg = "No tokenizer found for language '%s'" % (tag,)
            raise TokenizerNotFoundError(msg)
    # Given the language-specific tokenizer, we now build up the
    # end result as follows:
    #    * chunk the text using any given chunkers in turn
    #    * begin with basic whitespace tokenization
    #    * apply each of the given filters in turn
    #    * apply language-specific rules
    tokenizer = basic_tokenize
    if chunkers is not None:
        chunkers = list(chunkers)
        for i in range(len(chunkers) - 1, -1, -1):
            tokenizer = wrap_tokenizer(chunkers[i], tokenizer)
    if filters is not None:
        for f in filters:
            tokenizer = f(tokenizer)
    tokenizer = wrap_tokenizer(tokenizer, tk_func)
    return tokenizer


get_tokenizer._DOC_ERRORS = ["py", "py"]


class empty_tokenize(tokenize):  # noqa: N801
    """Tokenizer class that yields no elements."""

    _DOC_ERRORS = []

    def __init__(self):
        super().__init__("")

    def next(self):
        raise StopIteration()


class unit_tokenize(tokenize):  # noqa: N801
    """Tokenizer class that yields the text as a single token."""

    _DOC_ERRORS = []

    def __init__(self, text):
        super().__init__(text)
        self._done = False

    def next(self):
        if self._done:
            raise StopIteration()
        self._done = True
        return (self._text, 0)


class basic_tokenize(tokenize):  # noqa: N801
    """Tokenizer class that performs very basic word-finding.

    This tokenizer does the most basic thing that could work - it splits
    text into words based on whitespace boundaries, and removes basic
    punctuation symbols from the start and end of each word.
    """

    _DOC_ERRORS = []

    # Chars to remove from start/end of words
    strip_from_start = '"' + "'`(["
    strip_from_end = '"' + "'`]).!,?;:"

    def next(self):
        text = self._text
        offset = self._offset
        while True:
            if offset >= len(text):
                break
            # Find start of next word
            while offset < len(text) and text[offset].isspace():
                offset += 1
            s_pos = offset
            # Find end of word
            while offset < len(text) and not text[offset].isspace():
                offset += 1
            e_pos = offset
            self._offset = offset
            # Strip chars from font/end of word
            while s_pos < len(text) and text[s_pos] in self.strip_from_start:
                s_pos += 1
            while 0 < e_pos and text[e_pos - 1] in self.strip_from_end:
                e_pos -= 1
            # Return if word isn't empty
            if s_pos < e_pos:
                return (text[s_pos:e_pos], s_pos)
        raise StopIteration()


def _try_tokenizer(mod_name):
    """Look for a tokenizer in the named module.

    Returns the function if found, None otherwise.
    """
    mod_base = "enchant.tokenize."
    func_name = "tokenize"
    mod_name = mod_base + mod_name
    try:
        mod = __import__(mod_name, globals(), {}, func_name)
        return getattr(mod, func_name)
    except ImportError:
        return None


def wrap_tokenizer(tk1, tk2):
    """Wrap one tokenizer inside another.

    This function takes two tokenizer functions 'tk1' and 'tk2',
    and returns a new tokenizer function that passes the output
    of tk1 through tk2 before yielding it to the calling code.
    """
    # This logic is already implemented in the Filter class.
    # We simply use tk2 as the _split() method for a filter
    # around tk1.
    tkw = Filter(tk1)
    tkw._split = tk2
    return tkw


wrap_tokenizer._DOC_ERRORS = ["tk", "tk", "tk", "tk"]


class Chunker(tokenize):
    """Base class for text chunking functions.

    A chunker is designed to chunk text into large blocks of tokens.  It
    has the same interface as a tokenizer but is for a different purpose.
    """

    pass


class Filter:
    """Base class for token filtering functions.

    A filter is designed to wrap a tokenizer (or another filter) and do
    two things:

      * skip over tokens
      * split tokens into sub-tokens

    Subclasses have two basic options for customising their behaviour.  The
    method _skip(word) may be overridden to return True for words that
    should be skipped, and false otherwise.  The method _split(word) may
    be overridden as tokenization function that will be applied to further
    tokenize any words that aren't skipped.
    """

    def __init__(self, tokenizer):
        """Filter class constructor."""
        self._tokenizer = tokenizer

    def __call__(self, *args, **kwds):
        tkn = self._tokenizer(*args, **kwds)
        return self._TokenFilter(tkn, self._skip, self._split)

    def _skip(self, word):
        """Filter method for identifying skippable tokens.

        If this method returns true, the given word will be skipped by
        the filter.  This should be overridden in subclasses to produce the
        desired functionality.  The default behaviour is not to skip any words.
        """
        return False

    def _split(self, word):
        """Filter method for sub-tokenization of tokens.

        This method must be a tokenization function that will split the
        given word into sub-tokens according to the needs of the filter.
        The default behaviour is not to split any words.
        """
        return unit_tokenize(word)

    class _TokenFilter:
        """Private inner class implementing the tokenizer-wrapping logic.

        This might seem convoluted, but we're trying to create something
        akin to a meta-class - when Filter(tknzr) is called it must return
        a *callable* that can then be applied to a particular string to
        perform the tokenization.  Since we need to manage a lot of state
        during tokenization, returning a class is the best option.
        """

        _DOC_ERRORS = ["tknzr"]

        def __init__(self, tokenizer, skip, split):
            self._skip = skip
            self._split = split
            self._tokenizer = tokenizer
            # for managing state of sub-tokenization
            self._curtok = empty_tokenize()
            self._curword = ""
            self._curpos = 0

        def __iter__(self):
            return self

        def __next__(self):
            return self.next()

        def next(self):
            # Try to get the next sub-token from word currently being split.
            # If unavailable, move on to the next word and try again.
            while True:
                try:
                    (word, pos) = next(self._curtok)
                    return (word, pos + self._curpos)
                except StopIteration:
                    (word, pos) = next(self._tokenizer)
                    while self._skip(self._to_string(word)):
                        (word, pos) = next(self._tokenizer)
                    self._curword = word
                    self._curpos = pos
                    self._curtok = self._split(word)

        def _to_string(self, word):
            if type(word) is array.array:
                if word.typecode == "u":
                    return word.tounicode()
                elif word.typecode == "c":
                    return word.tostring()
            return word

        # Pass on access to 'offset' to the underlying tokenizer.
        def _get_offset(self):
            return self._tokenizer.offset

        def _set_offset(self, offset):
            msg = (
                "changing a tokenizers 'offset' attribute is deprecated;"
                " use the 'set_offset' method"
            )
            warnings.warn(msg, category=DeprecationWarning, stacklevel=2)
            self.set_offset(offset)

        offset = property(_get_offset, _set_offset)

        def set_offset(self, val, replaced=False):
            old_offset = self._tokenizer.offset
            self._tokenizer.set_offset(val, replaced=replaced)
            # If we move forward within the current word, also set on _curtok.
            # Otherwise, throw away _curtok and set to empty iterator.
            keep_curtok = True
            curtok_offset = val - self._curpos
            if old_offset > val:
                keep_curtok = False
            if curtok_offset < 0:
                keep_curtok = False
            if curtok_offset >= len(self._curword):
                keep_curtok = False
            if keep_curtok and not replaced:
                self._curtok.set_offset(curtok_offset)
            else:
                self._curtok = empty_tokenize()
                self._curword = ""
                self._curpos = 0


#  Pre-defined chunkers and filters start here


class URLFilter(Filter):
    r"""Filter skipping over URLs.
    This filter skips any words matching the following regular expression:

           ^[a-zA-Z]+:\/\/[^\s].*

    That is, any words that are URLs.
    """
    _DOC_ERRORS = ["zA"]
    _pattern = re.compile(r"^[a-zA-Z]+:\/\/[^\s].*")

    def _skip(self, word):
        if self._pattern.match(word):
            return True
        return False


class WikiWordFilter(Filter):
    r"""Filter skipping over WikiWords.
    This filter skips any words matching the following regular expression:

           ^([A-Z]\w+[A-Z]+\w+)

    That is, any words that are WikiWords.
    """
    _pattern = re.compile(r"^([A-Z]\w+[A-Z]+\w+)")

    def _skip(self, word):
        if self._pattern.match(word):
            return True
        return False


class EmailFilter(Filter):
    r"""Filter skipping over email addresses.
    This filter skips any words matching the following regular expression:

           ^.+@[^\.].*\.[a-z]{2,}$

    That is, any words that resemble email addresses.
    """
    _pattern = re.compile(r"^.+@[^\.].*\.[a-z]{2,}$")

    def _skip(self, word):
        if self._pattern.match(word):
            return True
        return False


class MentionFilter(Filter):
    r"""Filter skipping over @mention.
    This filter skips any words matching the following regular expression:

           (\A|\s)@(\w+)

    That is, any words that are @mention.
    """
    _DOC_ERRORS = ["zA"]
    _pattern = re.compile(r"(\A|\s)@(\w+)")

    def _skip(self, word):
        if self._pattern.match(word):
            return True
        return False


class HashtagFilter(Filter):
    r"""Filter skipping over #hashtag.
    This filter skips any words matching the following regular expression:

           (\A|\s)#(\w+)

    That is, any words that are #hashtag.
    """
    _DOC_ERRORS = ["zA"]
    _pattern = re.compile(r"(\A|\s)#(\w+)")

    def _skip(self, word):
        if self._pattern.match(word):
            return True
        return False


class HTMLChunker(Chunker):
    """Chunker for breaking up HTML documents into chunks of checkable text.

    The operation of this chunker is very simple - anything between a "<"
    and a ">" will be ignored.  Later versions may improve the algorithm
    slightly.
    """

    def next(self):
        text = self._text
        offset = self.offset
        while True:
            if offset >= len(text):
                break
            #  Skip to the end of the current tag, if any.
            if text[offset] == "<":
                maybe_tag = offset
                if self._is_tag(text, offset):
                    while text[offset] != ">":
                        offset += 1
                        if offset == len(text):
                            offset = maybe_tag + 1
                            break
                    else:
                        offset += 1
                else:
                    offset = maybe_tag + 1
            s_pos = offset
            #  Find the start of the next tag.
            while offset < len(text) and text[offset] != "<":
                offset += 1
            self._offset = offset
            # Return if chunk isn't empty
            if s_pos < offset:
                return (text[s_pos:offset], s_pos)
        raise StopIteration()

    def _is_tag(self, text, offset):
        if offset + 1 < len(text):
            if text[offset + 1].isalpha():
                return True
            if text[offset + 1] == "/":
                return True
        return False


# TODO: LaTeXChunker
